"""
A namedtuple representaton for the extracted info
"""

from datetime import datetime
from typing import Optional, NamedTuple, Dict, Any


Second = int


class Metadata(NamedTuple):
    """
    typically isn't used completely by one browser, includes
    partial information from browsers which supply the information
    """

    title: Optional[str] = None
    description: Optional[str] = None
    preview_image: Optional[str] = None
    duration: Optional[Second] = None

    @classmethod
    def make(
        cls,
        title: Optional[str] = None,
        description: Optional[str] = None,
        preview_image: Optional[str] = None,
        duration: Optional[Second] = None,
    ) -> Optional["Metadata"]:
        """
        Alternate constructor; only make the Metadata object if the user
        supplies at least one piece of data
        """
        if (title or description or preview_image or duration) is None:
            return None
        return cls(
            title=title,
            description=description,
            preview_image=preview_image,
            duration=duration,
        )


def test_make_metadata() -> None:
    assert Metadata.make(None, None, None, None) is None
    assert Metadata.make(title="webpage title", duration=5) is not None


class Visit(NamedTuple):
    url: str
    dt: datetime
    # hmm, does this being optional make it more annoying to consume
    # by other programs? reduces the amount of data that other programs
    # need to consume, so theres a tradeoff...
    metadata: Optional[Metadata] = None

    def serialize(self) -> Dict[str, Any]:
        return {
            "url": self.url,
            "dt": self.dt.timestamp(),
            "metadata": self.metadata._asdict() if self.metadata is not None else None,
        }



from datetime import datetime
from pathlib import Path
from typing import Optional, Type, Union

from sqlite_backup import sqlite_backup

from .log import logger
from .common import PathIsh, expand_path
from .browsers.all import DEFAULT_BROWSERS
from .browsers.common import Browser


def _progress(status: str, remaining: int, total: int) -> None:
    logger.debug(f"Copied {total-remaining} of {total} database pages...")


def _sqlite_backup(src: PathIsh, dest: PathIsh) -> None:
    logger.info(f"backing up {src} to {dest}")
    sqlite_backup(
        src,
        dest,
        wal_checkpoint=True,
        copy_use_tempdir=True,
        sqlite_backup_kwargs={"progress": _progress},
    )


def _path_backup(src: PathIsh, dest: PathIsh) -> Path:
    """
    User specified the exact path to backup a database, back it up
    """
    srcp: Path = expand_path(src)
    destp: Path = _default_pattern(srcp, dest)
    _sqlite_backup(srcp, destp)
    return destp


def _default_pattern(
    src: Path,
    to: PathIsh,
    browser_name: Optional[str] = None,
    pattern: Optional[str] = None,
) -> Path:
    """
    can pass a pattern with a format replacement field (for the timestamp)
    if you'd rather use a different format

    by default, this appends sqlite if thats not already the suffix,
    adds the name of the browser and a timestamp
    """
    to_p: Path = expand_path(to)
    assert to_p.is_dir(), f"{to_p} is not a directory!"
    suffix = src.suffix or ".sqlite"
    if pattern is None:
        pattern = (browser_name or "browser") + "-{}" + suffix
    # create pattern to timestamp backup filename
    now: str = datetime.utcnow().strftime("%Y%m%d%H%M%S")
    return to_p / pattern.format(now)


def backup_history(
    browser: Union[str, Type[Browser]],
    to: PathIsh,
    *,
    profile: str = "*",
    pattern: Optional[str] = None,
    save_type: Optional[str] = None,
) -> Path:
    """
    browser:
        typically the name of a browser the user provided from the CLI
        For library usage, user can pass a custom subclassed browser type which implements
            the required functions to locate the database
    to:
        path/string to backup database to
    profile:
        for browsers which have multiple profiles/users, this lets you specify
        a glob to select a particular profile
    pattern:
        pattern for the resulting timestamped filename, should include a format replacement field
    save_type:
        browser history, form history etc. If not supplied, assumes browser history
    """

    chosen: Type[Browser]
    browser_name: str
    if isinstance(browser, str):
        for extr in DEFAULT_BROWSERS:
            if browser.lower() == extr.__name__.lower():
                chosen = extr
                browser_name = browser.lower()
                break
        else:
            raise RuntimeError(f"Unknown browser: {browser}")
    else:
        chosen = browser
        browser_name = chosen.__name__.lower()
    src: Path
    if save_type == "form_history":
        src = chosen.locate_form_history(profile)
    else:
        src = chosen.locate_database(profile)
    dest = _default_pattern(src, to, browser_name, pattern)
    _sqlite_backup(src, dest)
    return dest



from .common import Paths, handle_path
from .firefox import Firefox

# seems to match firefox schema well enough for all of our usage


class Waterfox(Firefox):
    has_form_history_save = False

    @classmethod
    def data_directories(cls) -> Paths:
        return handle_path(
            {
                "linux": "~/.waterfox/",
                "darwin": "~/Library/Application Support/Waterfox/Profiles/",
            },
            browser_name=cls.__name__,
        )



from typing import Iterator, List, Optional, Type

from .common import PathIshOrConn
from .model import Visit
from .log import logger

from .browsers.common import Browser
from .browsers.all import DEFAULT_BROWSERS


def read_visits(
    path: PathIshOrConn, *, additional_browsers: Optional[List[Type[Browser]]] = None
) -> Iterator[Visit]:
    """
    Takes one sqlite database as input and returns 'Visit's
    """
    browsers: List[Type[Browser]] = additional_browsers or []
    browsers += DEFAULT_BROWSERS
    logger.info(f"Reading visits from {path}...")
    for br in browsers:
        if br.detect(path):
            logger.debug(f"Detected as {br.__name__}")
            yield from br.extract_visits(path)
            return
    else:
        raise RuntimeError(f"{path} didn't match any known schema")



from .common import Path, Paths


from .firefox import Firefox


class FirefoxMobile(Firefox):
    """
    New format, used on Android since 2020.
    The DB for stable Firefox is at /data/data/org.mozilla.firefox/files/places.sqlite
    The DB for Fenix is at /data/data/org.mozilla.fenix/files/places.sqlite

    Note that currently the DB schema seems more or less identical, except for timestamps:
    they are in microseconds on desktop and in milliseconds on mobile.
    This is handled in base Firefox class for both cases.
    """

    has_form_history_save = False

    # unclear how reliable it is
    # but we prefer to set it anyway to tell apart whether visits came from desktop or mobile
    # see https://github.com/seanbreckenridge/browserexport/issues/14#issuecomment-1037891476
    detector = "SELECT * FROM moz_meta, moz_tags"
    has_save = False

    @classmethod
    def data_directories(cls) -> Paths:
        raise NotImplementedError("Only available on Android")

    @classmethod
    def locate_database(cls, profile: str = "*") -> Path:
        raise NotImplementedError("Only available on Android")



from .common import (
    Path,
    Schema,
    Iterator,
    Visit,
    Browser,
    Metadata,
    unquote,
    from_datetime_microseconds,
    execute_query,
    handle_path,
    handle_glob,
    Paths,
    PathIshOrConn,
    sqlite3,
    logger,
)


class Palemoon(Browser):
    """
    uses a similar directory structure to firefox,
    but has enough things different that worth subclassing Browser
    """

    detector = "moz_historyvisits"

    @classmethod
    def detect(cls, path: PathIshOrConn) -> bool:
        # if this doesnt have the moz_historyvisits, exit
        if not super().detect(path):
            return False
        try:
            # Palemoon doesn't have the moz_meta table, so can use that
            # to make sure this is palemoon and not some other firefox derivative?
            logger.debug(
                "May be Palemoon, running query on moz_meta to ensure this isn't another Firefox derivative"
            )
            list(execute_query(path, "Select * FROM moz_meta"))
            logger.debug("'moz_meta' query failed, not Palemoon")
            return False
        except sqlite3.OperationalError:
            logger.debug(
                "moz_historyvisits exists but moz_meta doesn't, detected as Palemoon"
            )
            return True

    # seems to store less info that firefox schema
    # no description or preview_image
    schema = Schema(
        cols=[
            "P.url",
            "V.visit_date",
            "P.title",
        ],
        where="FROM moz_historyvisits as V, moz_places as P WHERE V.place_id = P.id",
        order_by="V.visit_date",
    )

    @classmethod
    def extract_visits(cls, path: PathIshOrConn) -> Iterator[Visit]:
        for row in execute_query(path, cls.schema.query):
            yield Visit(
                url=unquote(row["url"]),
                dt=from_datetime_microseconds(row["visit_date"]),
                metadata=Metadata.make(title=row["title"]),
            )

    # seems the non-linux community is pretty small?
    # https://forum.palemoon.org/viewforum.php?f=41
    # no easy to way to install except to build from source
    # if someone is actually using this on mac, feel free to make an issue
    @classmethod
    def data_directories(cls) -> Paths:
        return handle_path(
            {
                "linux": "~/.moonchild productions/pale moon/",
            },
            browser_name=cls.__name__,
        )

    @classmethod
    def locate_database(cls, profile: str = "*") -> Path:
        dd = cls.data_directories()
        return handle_glob(dd, profile + "/places.sqlite")



from .common import (
    Iterator,
    Visit,
    Metadata,
    PathIshOrConn,
    Browser,
    unquote,
    Schema,
    Path,
    datetime,
    timezone,
    handle_glob,
    handle_path,
    execute_query,
    Paths,
)

WINDOWS_EPOCH_OFFSET = 11644473600


def _chrome_date_to_utc(chrome_time: int) -> datetime:
    ts = (chrome_time / 1_000_000) - WINDOWS_EPOCH_OFFSET
    return datetime.fromtimestamp(ts, tz=timezone.utc)


class Chrome(Browser):
    detector = "keyword_search_terms"

    schema = Schema(
        cols=[
            "U.url",
            "U.title",
            "V.visit_time",
            "V.visit_duration",
        ],
        where="FROM visits as V, urls as U WHERE V.url = U.id",
        order_by="V.visit_time",
    )

    @classmethod
    def extract_visits(cls, path: PathIshOrConn) -> Iterator[Visit]:
        for row in execute_query(path, cls.schema.query):
            dur = int(row["visit_duration"])
            yield Visit(
                url=unquote(row["url"]),
                dt=_chrome_date_to_utc(row["visit_time"]),
                metadata=Metadata.make(
                    title=row["title"],
                    duration=None if dur == 0 else dur // 1_000_000,
                ),
            )

    @classmethod
    def data_directories(cls) -> Paths:
        return handle_path(
            {
                "linux": "~/.config/google-chrome/",
                "darwin": "~/Library/Application Support/Google/Chrome/",
            },
            browser_name=cls.__name__,
        )

    @classmethod
    def locate_database(cls, profile: str = "*") -> Path:
        dd = cls.data_directories()
        return handle_glob(dd, profile + "/History")



from collections import Counter
from typing import List
from pprint import pprint

from urllib.parse import urlparse

from .model import Visit


def demo_visit(visits: List[Visit]) -> None:
    print("Demo: Your most common sites....")
    pprint(Counter(map(lambda v: urlparse(v.url).netloc, visits)).most_common(10))



from .common import (
    handle_path,
    Paths,
)


from .chrome import Chrome


class Chromium(Chrome):
    @classmethod
    def data_directories(cls) -> Paths:
        return handle_path(
            {
                "linux": ("~/.config/chromium/", "~/snap/chromium/common/chromium/"),
                "darwin": "~/Library/Application Support/Chromium/",
            },
            browser_name=cls.__name__,
        )



import sqlite3

from typing import Iterator

from .common import expand_path, PathIshOrConn


def _execute_conn(conn: sqlite3.Connection, query: str) -> Iterator[sqlite3.Row]:
    """
    Given an open sqlite3 connection, execute a query
    """
    conn.row_factory = sqlite3.Row
    conn.text_factory = lambda b: b.decode(errors="ignore")
    for row in conn.execute(query):
        yield row


def execute_query(path: PathIshOrConn, query: str) -> Iterator[sqlite3.Row]:
    """
    Given a str, path, or sqlite3 connection, execute a query
    """
    if isinstance(path, sqlite3.Connection):
        # WARNING: if this happens but you're yielding lazily from a function
        # the connection might close before this query finishes executing
        yield from _execute_conn(path, query)
    else:
        p: str = str(expand_path(path))
        with sqlite3.connect(f"file:{p}?immutable=1", uri=True) as c:
            yield from _execute_conn(c, query)



from typing import Union, TypeVar, Callable, Optional
from pathlib import Path
from sqlite3 import Connection

PathIsh = Union[str, Path]
# a path or a connection to a database
PathIshOrConn = Union[PathIsh, Connection]


def expand_path(path: PathIsh) -> Path:
    if isinstance(path, str):
        path = Path(path)
    return path.expanduser().absolute()


T = TypeVar("T")


def func_if_some(maybe: Optional[T], func: Callable[[T], T]) -> Optional[T]:
    """if 'maybe' is not None, run the specified function"""
    if maybe is not None:
        return func(maybe)
    return maybe



from pathlib import Path
from setuptools import setup, find_namespace_packages  # type: ignore[import]


long_description = Path("README.md").read_text()
reqs = Path("requirements.txt").read_text().strip().splitlines()

pkg = "browserexport"
setup(
    name=pkg,
    version="0.2.5",
    url="https://github.com/seanbreckenridge/browserexport",
    author="Sean Breckenridge",
    author_email="seanbrecke@gmail.com",
    description=(
        """save and merge browser history and metadata from different browsers"""
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    packages=find_namespace_packages(include=[f"{pkg}*"]),
    package_data={pkg: ["py.typed"]},
    python_requires=">=3.7",
    install_requires=reqs,
    extras_require={
        "testing": [
            "pytest",
            "mypy",
            "flake8",
        ],
    },
    keywords="firefox history backup data",
    entry_points={"console_scripts": ["browserexport = browserexport.__main__:cli"]},
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
)



import sys
import logging
import json as jsn
from typing import List, Optional, Sequence, Iterator

import click

from .browsers.all import DEFAULT_BROWSERS

CONTEXT_SETTINGS = {
    "max_content_width": 110,
    "show_default": True,
}


# target for python3 -m browserexport and console_script using click
@click.group(context_settings=CONTEXT_SETTINGS)
@click.option("--debug", is_flag=True, default=False, help="Increase log verbosity")
def cli(debug: bool) -> None:
    """
    Backup and merge your browser history!
    """
    from .log import setup

    if debug:
        setup(logging.DEBUG)


browsers_have_save = [b.__name__.lower() for b in DEFAULT_BROWSERS if b.has_save]

browsers_have_form_save = [
    b.__name__.lower() for b in DEFAULT_BROWSERS if b.has_form_history_save
]

# define click options
profile = click.option(
    "-p",
    "--profile",
    type=str,
    default="*",
    help="Use to pick the correct profile to back up. If unspecified, will assume a single profile",
)

path = click.option(
    "--path",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Specify a direct path to a database to back up",
)

store_to = click.option(
    "-t",
    "--to",
    type=click.Path(exists=True, file_okay=False),
    required=True,
    help="Directory to store backup to",
)

print_json = click.option(
    "-j",
    "--json",
    is_flag=True,
    default=False,
    required=False,
    help="Print result to STDOUT as JSON",
)

stream_json = click.option(
    "-s",
    "--stream",
    is_flag=True,
    default=False,
    required=False,
    help="Stream JSON objects instead of printing a JSON list",
)


@cli.command()
@click.option(
    "-b",
    "--browser",
    type=click.Choice(browsers_have_save, case_sensitive=False),
    required=False,
    help="Browser name to backup history for",
)
@click.option(
    "--form-history",
    type=click.Choice(browsers_have_form_save, case_sensitive=False),
    required=False,
    help="Browser name to backup form (input field) history for",
)
@profile
@path
@store_to
@click.pass_context
def save(
    ctx: click.Context,
    browser: Optional[str],
    form_history: Optional[str],
    profile: str,
    to: str,
    path: Optional[str],
) -> None:
    """
    Backs up a current browser database file
    """
    from .save import backup_history, _path_backup

    if path is not None:
        _path_backup(path, to)
    elif form_history is not None:
        backup_history(form_history, to, profile=profile, save_type="form_history")
    elif browser is not None:
        backup_history(browser, to, profile=profile)
    else:
        click.secho(
            "Error: must provide one of '--browser', '--form-history', or '--path'",
            err=True,
            fg="red",
        )
        click.echo(ctx.get_help())


def _handle_merge(dbs: List[str], *, json: bool, stream: bool) -> None:
    from .model import Visit
    from .merge import read_and_merge

    ivis: Iterator[Visit] = read_and_merge(dbs)
    if json or stream:
        if stream:
            for v in ivis:
                sys.stdout.write(jsn.dumps(v.serialize()))
            sys.stdout.flush()
        else:
            click.echo(jsn.dumps([v.serialize() for v in ivis]))
    else:
        import IPython  # type: ignore[import]

        from .demo import demo_visit

        vis: List[Visit] = list(ivis)
        demo_visit(vis)
        header = f"Use {click.style('vis', fg='green')} to access visit data"

        IPython.embed(header=header)


@cli.command()
@click.argument(
    "SQLITE_DB", type=click.Path(exists=True, dir_okay=False), required=True
)
@stream_json
@print_json
def inspect(sqlite_db: str, json: bool, stream: bool) -> None:
    """
    Extracts visits from a single sqlite database

    \b
    Provide a history database as the first argument
    Drops you into a REPL to access the data
    """
    _handle_merge([sqlite_db], json=json, stream=stream)


@cli.command()
@click.argument(
    "SQLITE_DB", type=click.Path(exists=True, dir_okay=False), nargs=-1, required=True
)
@stream_json
@print_json
def merge(sqlite_db: Sequence[str], json: bool, stream: bool) -> None:
    """
    Extracts visits from multiple sqlite databases

    \b
    Provide multiple sqlite databases as positional arguments, e.g.:
    browserexport merge ~/data/firefox/*.sqlite

    Drops you into a REPL to access the data
    """
    _handle_merge(list(sqlite_db), json=json, stream=stream)


if __name__ == "__main__":
    cli(prog_name="browserexport")



"""
Setup logging for this module
"""

import os
import logging

from typing import Optional

from logzero import setup_logger  # type: ignore[import]

DEFAULT_LEVEL = logging.WARNING

# global access to the logger
logger: logging.Logger


def _monkey_patch_sqlite_backup_logs(level: int) -> None:
    import sqlite_backup.log

    sqlite_backup.log.setup(level)


# logzero handles adding handling/modifying levels fine
# can be imported/configured multiple times
def setup(level: Optional[int] = None) -> logging.Logger:
    chosen_level = level or int(os.environ.get("BROWSEREXPORT_LOGS", DEFAULT_LEVEL))
    lgr: logging.Logger = setup_logger(name=__package__, level=chosen_level)
    _monkey_patch_sqlite_backup_logs(chosen_level)
    return lgr


# runs the first time this file is run, setup can be imported/run multiple times in other places
logger = setup()



from .common import Paths, handle_path


from .chrome import Chrome


class Brave(Chrome):
    @classmethod
    def data_directories(cls) -> Paths:
        return handle_path(
            {
                "linux": "~/.config/BraveSoftware/Brave-Browser/",
                "darwin": "~/Library/Application Support/BraveSoftware/Brave-Browser/",
            },
            browser_name=cls.__name__,
        )



from typing import List, Type
from .common import Browser
from .firefox import Firefox
from .waterfox import Waterfox
from .palemoon import Palemoon
from .chrome import Chrome
from .chromium import Chromium
from .brave import Brave
from .safari import Safari
from .vivaldi import Vivaldi
from .firefox_mobile import FirefoxMobile
from .firefox_mobile_legacy import FirefoxMobileLegacy

# As this is a namespace package, you're free to add additional files
# to this package in a separate directory, and then append them (or override this file, by
# placing your namespace package before this on your PYTHONPATH),
# to this list, and they'd be added to the global list in the rest of the library
#
# for more information, see:
# https://www.python.org/dev/peps/pep-0420/#dynamic-path-computation
# https://packaging.python.org/guides/creating-and-discovering-plugins/#using-namespace-packages
# https://packaging.python.org/guides/packaging-namespace-packages/
# https://github.com/seanbreckenridge/reorder_editable

DEFAULT_BROWSERS: List[Type[Browser]] = [
    Chrome,
    Firefox,
    Safari,
    Brave,
    Waterfox,
    Chromium,
    Vivaldi,
    Palemoon,  # has to be after waterfox/firefox derivates, else it could be mis-detected
    FirefoxMobile,
    FirefoxMobileLegacy,
]



from .common import (
    Iterator,
    Visit,
    Metadata,
    PathIshOrConn,
    Browser,
    unquote,
    Path,
    Schema,
    execute_query,
    from_datetime_microseconds,
    Paths,
)


class FirefoxMobileLegacy(Browser):
    """
    Legacy format, used on Android pre-2020.
    The DB was at /data/data/org.mozilla.firefox/files/places.sqlite
    """

    detector = "remote_devices"
    schema = Schema(
        cols=[
            "H.url",
            "V.date",
            "H.title",
        ],
        where="FROM visits as V, history as H WHERE V.history_guid = H.guid",
        order_by="V.date",
        # todo: bookmarks, searchhistory tables might be interesting
    )
    has_save = False

    @classmethod
    def extract_visits(cls, path: PathIshOrConn) -> Iterator[Visit]:
        for row in execute_query(path, cls.schema.query):
            yield Visit(
                url=unquote(row["url"]),
                dt=from_datetime_microseconds(row["date"]),
                metadata=Metadata.make(
                    title=row["title"],
                ),
            )

    @classmethod
    def data_directories(cls) -> Paths:
        raise NotImplementedError("Only available on Android")

    @classmethod
    def locate_database(cls, profile: str = "*") -> Path:
        raise NotImplementedError("Only available on Android")



from .common import (
    Iterator,
    Visit,
    Metadata,
    PathIshOrConn,
    Browser,
    Schema,
    Path,
    unquote,
    datetime,
    timezone,
    execute_query,
    handle_glob,
    handle_path,
    Paths,
)

# Referenced:
# https://github.com/karlicoss/promnesia/blob/0e1e9a1ccd1f07b2a64336c18c7f41ca24fcbcd4/src/promnesia/sources/browser.py#L222
# https://web.archive.org/web/20201026130310/http://fileformats.archiveteam.org/wiki/History.db


def _safari_date_to_utc(safari_time: int) -> datetime:
    ts = safari_time + 978307200
    return datetime.fromtimestamp(ts, tz=timezone.utc)


class Safari(Browser):
    detector = "history_tombstones"

    schema = Schema(
        cols=[
            "U.url",
            "V.visit_time",
            "V.title",
        ],
        where="FROM history_visits as V, history_items as U WHERE V.history_item = U.id",
        order_by="V.visit_time",
    )

    @classmethod
    def extract_visits(cls, path: PathIshOrConn) -> Iterator[Visit]:
        for row in execute_query(path, cls.schema.query):
            yield Visit(
                url=unquote(row["url"]),
                dt=_safari_date_to_utc(row["visit_time"]),
                metadata=Metadata.make(title=row["title"]),
            )

    @classmethod
    def data_directories(cls) -> Paths:
        return handle_path(
            {
                "darwin": "~/Library/Safari/",
            },
            browser_name=cls.__name__,
            default_behaviour="mac",
        )

    @classmethod
    def locate_database(cls, profile: str = "*") -> Path:
        dd = cls.data_directories()
        return handle_glob(dd, profile + "History.db")



"""
Merges multiple history sqlite databases into one
"""

import warnings
from datetime import datetime
from typing import Iterator, Iterable, Sequence, Set, Tuple, List

from .log import logger
from .model import Visit
from .common import PathIsh, expand_path
from .parse import read_visits

# not sure on the typing/Sequence's with splat here
# works fine though, each of these accept variadic arguments
# with either PathIsh-things or Iterator/List things w/ Visits


def read_and_merge(paths: Sequence[PathIsh]) -> Iterator[Visit]:
    """
    Receives any amount of Path-like databases as input,
    reads Visits from each of those databases,
    and merges them together (removing duplicates)
    """
    pths = [expand_path(p) for p in paths]
    hst: List[Iterator[Visit]] = list(map(read_visits, pths))
    yield from merge_visits(hst)


def merge_visits(sources: Sequence[Iterable[Visit]]) -> Iterator[Visit]:
    """
    Removes duplicate Visit items from multiple sources
    """
    if len(sources) == 0:
        warnings.warn("merge_visits received no sources!")
    else:
        logger.debug(f"merging information from {len(sources)} sources...")
    # use combination of URL, visit date and visit type to uniquely identify visits
    emitted: Set[Tuple[str, datetime]] = set()
    duplicates = 0
    for src in sources:
        for vs in src:
            key = (vs.url, vs.dt)
            if key in emitted:
                # logger.debug(f"skipping {key} => {vs}")
                duplicates += 1
                continue
            yield vs
            emitted.add(key)
    logger.debug("Summary: removed {} duplicates...".format(duplicates))
    logger.info("Summary: returning {} visit entries...".format(len(emitted)))



from .common import (
    Iterator,
    Visit,
    Metadata,
    PathIshOrConn,
    Browser,
    unquote,
    Path,
    Schema,
    execute_query,
    from_datetime_microseconds,
    func_if_some,
    handle_glob,
    handle_path,
    Paths,
)


class Firefox(Browser):
    detector = "SELECT * FROM moz_meta, moz_annos"
    schema = Schema(
        cols=[
            "P.url",
            # Hack to tell apart whether timestamp is stored in microseconds (on desktop) or in milliseconds (on mobile)
            # We set 300_000_000 as threshold, it's year 1979, so definitely before Firefox existed,
            # and the same multipied by 1000 is year 11476, also enough time for us not to care.
            "(CASE WHEN (V.visit_date > 300000000 * 1000000) THEN V.visit_date ELSE V.visit_date * 1000 END) AS visit_date",
            "V.visit_date",
            "P.title",
            "P.description",
            "P.preview_image_url",
        ],
        where="FROM moz_historyvisits as V, moz_places as P WHERE V.place_id = P.id",
        order_by="V.visit_date",
    )

    @classmethod
    def extract_visits(cls, path: PathIshOrConn) -> Iterator[Visit]:
        for row in execute_query(path, cls.schema.query):
            yield Visit(
                url=unquote(row["url"]),
                dt=from_datetime_microseconds(row["visit_date"]),
                metadata=Metadata.make(
                    title=row["title"],
                    description=row["description"],
                    preview_image=func_if_some(row["preview_image_url"], unquote),
                ),
            )

    @classmethod
    def data_directories(cls) -> Paths:
        return handle_path(
            {
                "linux": ("~/.mozilla/firefox/", "~/snap/firefox/common/.mozilla/firefox/"),
                "darwin": "~/Library/Application Support/Firefox/Profiles/",
            },
            browser_name=cls.__name__,
        )

    @classmethod
    def locate_database(cls, profile: str = "*") -> Path:
        dd = cls.data_directories()
        return handle_glob(dd, profile + "/places.sqlite")

    has_form_history_save = True

    @classmethod
    def locate_form_history(cls, profile: str) -> Path:
        dd = cls.data_directories()
        return handle_glob(dd, profile + "/formhistory.sqlite")



# ununsed imports here, to bring them into scope for other files
import sys
import sqlite3
import warnings

from itertools import chain
from pathlib import Path
from urllib.parse import unquote
from datetime import datetime, timezone
from typing import List, Iterator, Optional, NamedTuple, Dict, Union, Tuple, Sequence
from dataclasses import dataclass


from ..log import logger
from ..model import Visit, Metadata
from ..common import PathIsh, PathIshOrConn, func_if_some, expand_path
from ..sqlite import execute_query


@dataclass
class Schema:
    cols: List[str]
    where: str
    order_by: Optional[str] = None

    @property
    def query(self) -> str:
        qr = f"SELECT {', '.join(self.cols)} {self.where}"
        if self.order_by is not None:
            qr += f" ORDER BY {self.order_by}"
        return qr


Detector = str
Paths = Sequence[Path]


@dataclass
class Browser:
    schema: Schema  # used to create the query to extract visit from database
    detector: Detector  # semi-unique name of table, or a query to run on database to detect this type
    has_save: bool = True  # if this browser works with the save command
    has_form_history_save: bool = False  # if this can backup form history

    @classmethod
    def detect(cls, path: PathIshOrConn) -> bool:
        """
        Run the detector against the given path/connection to detect if the current Browser matches the schema
        """
        # if the user provided something that had spaces - is a query
        if " " in cls.detector.strip():
            detector_query = cls.detector
        else:
            detector_query = f"SELECT * FROM {cls.detector}"
        logger.debug(f"{cls.__name__}: Running detector query '{detector_query}'")
        try:
            list(execute_query(path, detector_query))
            return True
        except sqlite3.OperationalError as sql_err:
            logger.debug(str(sql_err))
            return False

    @classmethod
    def extract_visits(cls, path: PathIshOrConn) -> Iterator[Visit]:
        """
        Given a path or a sqlite3 connection, extract visits
        """
        raise NotImplementedError

    @classmethod
    def data_directory(cls) -> Path:
        warnings.warn("'data_directory' method is deprecated. Please switch to 'data_directories'")
        dirs = cls.data_directories()
        assert len(dirs) > 0
        if len(dirs) > 1:
            logger.warn('got multiple alternaitves for data directory, picking the first')
        return dirs[0]

    @classmethod
    def data_directories(cls) -> Paths:
        """
        The local data directories for this browser
        """
        raise NotImplementedError

    @classmethod
    def locate_database(cls, profile: str) -> Path:
        """
        Locate this database on the users' computer so it can be backed up
        """
        raise NotImplementedError

    @classmethod
    def locate_form_history(cls, profile: str) -> Path:
        """
        Locate the Form History (e.g. Usernames/Form fields) database so it can be backed up
        """
        raise NotImplementedError


def from_datetime_microseconds(ts: int) -> datetime:
    return datetime.fromtimestamp(ts / 1_000_000, tz=timezone.utc)


errmsg = """Expected to match single database, got {}.
You can use the --profile argument to select one of the profiles/match a particular file"""


def handle_glob(bases: Sequence[Path], stem: str, recursive: bool = False) -> Path:
    dbs: List[Path]
    method = Path.rglob if recursive else Path.glob
    dbs = list(chain(*[method(base, stem) for base in bases]))
    recur_desc = "recursive" if recursive else "non recursive"
    logger.debug(f"Glob {bases} with {stem} ({recur_desc}) matched {dbs}")
    if len(dbs) > 1:
        raise RuntimeError(errmsg.format(dbs))
    elif len(dbs) == 1:
        # found the match!
        return dbs[0]
    else:
        # if we werent trying to search recursively, try a recursive search as a fallback
        if not recursive:
            return handle_glob(bases, stem, recursive=True)
        else:
            raise RuntimeError(f"Could not find database, using '{bases}' and '{stem}'")


PathMapEntry = Union[PathIsh, Sequence[PathIsh]]


def handle_path(
    pathmap: Dict[str, PathMapEntry],
    browser_name: str,
    *,
    key: Optional[str] = None,
    default_behaviour: str = "linux",
) -> Paths:
    """
    Handles the repetitive task of having to resolve/expand a path
    which describes the location of the data directory on each
    opreating system
    """
    loc: Sequence[PathIsh]
    # if the user didn't provide a key, assume this is a 'sys.platform' map - using
    # darwin/linux to specify the location
    if key is None:
        key = sys.platform
    # use the key provided, or the first item (dicts after python3.7 are ordered)
    # in the pathmap if that doesnt exist
    maybeloc: Optional[PathMapEntry] = pathmap.get(key)
    if maybeloc is None:
        warnings.warn(
            f"""Not sure where {browser_name} history is installed on {sys.platform}
Defaulting to {default_behaviour} behaviour...

If you're using a browser/platform this currently doesn't support, please make an issue
at https://github.com/seanbreckenridge/browserexport/issues/new with information.
In the meantime, you can point this directly at a history database using the --path flag"""
        )
        maybeloc = pathmap[list(pathmap.keys())[0]]
    assert maybeloc is not None  # convince mypy
    if isinstance(maybeloc, (Path, str)):
        loc = [maybeloc]
    else:
        loc = maybeloc
    return tuple(expand_path(p) for p in loc)


def test_handle_path() -> None:

    import pytest
    import sys

    oldplatform = sys.platform

    sys.platform = "linux"

    from .firefox import Firefox

    expected_linux = (
        Path("~/.mozilla/firefox/").expanduser().absolute(),
        Path("~/snap/firefox/common/.mozilla/firefox/").expanduser().absolute(),
    )

    assert Firefox.data_directories() == expected_linux

    sys.platform = "darwin"
    assert Firefox.data_directories() == (
        Path("~/Library/Application Support/Firefox/Profiles/").expanduser().absolute(),
    )

    sys.platform = "something else"
    # should default to linux
    with pytest.warns(UserWarning, match=r"history is installed"):
        assert Firefox.data_directories() == expected_linux

    sys.platform = oldplatform
